#include <avr/io.h>
#include "debug.h"
#include <avr/interrupt.h>
#include "simpleio.h"
#include <stdbool.h>

#define SPI_SS   IOPIN(D, 2)
#define SPI_MISO IOPIN(B, 4)

#define DBGO IOPIN(B, 0)

typedef enum
{
	Flash_Opcode_Invalid       = 0,
	Flash_Opcode_ReadContArray = 0xE8,
	Flash_Opcode_WriteBuff1    = 0x82,
	Flash_Opcode_ErasePage     = 0x81
}	Flash_Opcode_t;

typedef enum
{
	Flash_Area_None     = 0,
	Flash_Area_Hello    = 1, // 0x000000 ... 0x000003
	Flash_Area_Sysinfo  = 2, // 0x00000F ... 0x00006F
	Flash_Area_Channels = 3, // 0x000200 ... 0x00043F
	Flash_Area_Data     = 4, // 0x000500 ... 0x0FFFFF
}	Flash_Area_t;

typedef enum
{
	TrxFlags_SpiBusy =         (1<<0),
	TrxFlags_SentHello =       (1<<1),
	TrxFlags_EraseLogData =    (1<<2),
	TrxFlags_ReceivedLogData = (1<<3),
}	TrxFlags_t;

volatile uint8_t trx_flags = 0; //TrxFlags

volatile Flash_Opcode_t flash_opcode = Flash_Opcode_Invalid;
volatile uint8_t spi_bytenum = 0;

// yes, it's quite crap to have this data in ram, but it's fast.
volatile uint8_t flash_data_hello[3] = { 0xAA, 0x55, 0x01 };
volatile uint8_t flash_data_log[256] = { 0xFF };


volatile union
{
	struct
	{
		uint16_t lower;
		uint8_t upper;
	} val24;
	uint8_t val8[3];
} flash_address;

ISR(SPI_STC_vect)
{
	IOPIN_H(DBGO);
	uint8_t d = SPDR;
	if(spi_bytenum == 0)
	{
		// byte 0 in transfer is the opcode
		SPDR = 0x00;
		flash_opcode = (Flash_Opcode_t)d;
	}
	else if(spi_bytenum < 4)
	{
		SPDR = 0x00;
		// byte 1 ... 3 are address bytes
		flash_address.val8[sizeof(flash_address) - spi_bytenum] = d;
		if(spi_bytenum == 3)
		{
			// command completely received, now do what needs to be done
			//uart_debug_write("%2x:%4x", flash_opcode, flash_address.val24.lower);
		}
	}
	else if(flash_opcode == Flash_Opcode_ReadContArray && spi_bytenum >= 7)
	{
		
		// when reading from the flash, we have to start to transmit
		// the data already one byte earlier due to delay of MISO
		if(flash_address.val24.lower <= 0x0003)
		{
			// hello bytes
			SPDR = flash_data_hello[flash_address.val24.lower & 0x03];
		}
		else if(flash_address.val24.lower >= 0x0500)
		{
			// logging data section
			SPDR = flash_data_log[flash_address.val24.lower & 0xFF];
		}
		
		flash_address.val24.lower++;
	}
	else if(flash_opcode == Flash_Opcode_WriteBuff1 && spi_bytenum >= 4)
	{
		if(flash_address.val24.lower >= 0x0500)
		{
			flash_data_log[flash_address.val24.lower & 0xFF] = d;
		}
		
		flash_address.val24.lower++;
	}
	else
	{
		// when there's no specific handler for the transmission, reply 0x00
		SPDR = 0x00;
	}
	
	// byte number is only important for the first n bytes, after that
	// there are anyway only data bytes (which are handled separately)
	if(spi_bytenum < 0xFF)
		spi_bytenum++;
	IOPIN_L(DBGO);
}



ISR(PCINT0_vect)
{
	if(IOPIN_V(SPI_SS) == 0)
	{
		spi_bytenum = 0;
		flash_opcode = 0;
		//preload SPI data register
		SPDR = 0x00;
		trx_flags |= TrxFlags_SpiBusy;
		//uart_debug_putc(trx_flags);
	}
	else
	{
		trx_flags &= ~(TrxFlags_SpiBusy);
		// transmission of a packet is done, what has been done?
		if(flash_opcode == Flash_Opcode_ErasePage)
		{
			trx_flags |= TrxFlags_EraseLogData;	
		}
		else if(Flash_Opcode_ReadContArray)
		{
			
			// the 3 here looks very odd, but the actual byte has to be in 
			// SPDR one byte before it's beint transmitted and also the
			// flash address counter will increment after the byte is read
			if(flash_address.val24.lower == 0x0003)
			{
				trx_flags |= TrxFlags_SentHello;
			}
		}
		else if(Flash_Opcode_WriteBuff1)
		{
			if(flash_address.val24.lower >= 0x0500)
			{
				trx_flags |= TrxFlags_ReceivedLogData;
			}
			
		}
		//uart_debug_putc(trx_flags);
	}
	
		
	
}


int main(void)
{
	uart_debug_init();
	
	IOPIN_O(SPI_MISO);
	//enable SPI slave
	SPCR = (1<<SPIE) | (1<<SPE) | (0<<DORD) | (0<<MSTR) | (0<<CPOL) | (0<<CPHA);
	
	PCMSK0 |= (1<<PCINT2);
	PCICR |= (1<<PCIE0);
	
	IOPIN_O(DBGO);
	IOPIN_L(DBGO);
	
	sei();
	
	uart_debug_writeln("oh, hai! this should run through the FIFO.");
	
	//uint8_t trx_flags_old = trx_flags;
    
    while (1) 
    {
		
		if(trx_flags & TrxFlags_SentHello)
		{
			trx_flags &= ~TrxFlags_SentHello;
			uart_debug_writeln("Hello!");
		}
		if(trx_flags & TrxFlags_EraseLogData)
		{
			trx_flags &= ~TrxFlags_EraseLogData;
			uart_debug_writeln("we will have to erase log data");
		}
		if(trx_flags & TrxFlags_ReceivedLogData)
		{
			trx_flags &= ~TrxFlags_ReceivedLogData;
			uart_debug_writeln("we received log data @%2x", (flash_address.val24.lower & 0xFF) - 65);
		}
			
		
		
		//trx_flags_old = trx_flags;
		
    }
}

